#ifndef WIN32
#include <errno.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#endif
#include <msp/core/systemerror.h>
#include "console.h"
#include "handle_private.h"

using namespace std;

namespace {

#ifndef WIN32
termios orig_attr;
#endif

} // namespace

namespace Msp {
namespace IO {

Console::Console(unsigned n)
{
	if(n>2)
		throw invalid_argument("Console::Console");

	mode = (n==0 ? M_READ : M_WRITE);

#ifdef WIN32
	switch(n)
	{
	case 0: *handle = GetStdHandle(STD_INPUT_HANDLE); break;
	case 1: *handle = GetStdHandle(STD_OUTPUT_HANDLE); break;
	case 2: *handle = GetStdHandle(STD_ERROR_HANDLE); break;
	}
#else
	*handle = n;

	if(n==0)
		tcgetattr(*handle, &orig_attr);
#endif

	if(n==0)
		set_events(P_INPUT);
}

Console::~Console()
{
#ifndef WIN32
	if(handle==0)
		tcsetattr(*handle, TCSADRAIN, &orig_attr);
#endif
}

void Console::set_block(bool b)
{
#ifdef WIN32
	// XXX Dunno how to do this in win32
	(void)b;
#else
	int flags = fcntl(*handle, F_GETFL);
	flags = (flags&~O_NONBLOCK) | (b?0:O_NONBLOCK);
	fcntl(*handle, F_SETFL, flags);
#endif
}

void Console::set_local_echo(bool e)
{
	check_access(M_READ);

#ifdef WIN32
	DWORD m;
	GetConsoleMode(*handle, &m);
	SetConsoleMode(*handle, (m&~ENABLE_ECHO_INPUT) | (e?ENABLE_ECHO_INPUT:0));
#else
	termios t;
	tcgetattr(*handle, &t);
	t.c_lflag = (t.c_lflag&~ECHO) | (e?ECHO:0);
	tcsetattr(*handle, TCSADRAIN, &t);
#endif
}

void Console::set_line_buffer(bool l)
{
	check_access(M_READ);

#ifdef WIN32
	DWORD m;
	GetConsoleMode(*handle, &m);
	SetConsoleMode(*handle, (m&~ENABLE_LINE_INPUT) | (l?ENABLE_LINE_INPUT:0));
#else
	// XXX ICANON does more than just set line buffering, may need a bit more thought
	termios t;
	tcgetattr(*handle, &t);
	t.c_lflag = (t.c_lflag&~ICANON) | (l?ICANON:0);
	tcsetattr(*handle, TCSADRAIN, &t);
#endif
}

void Console::get_size(unsigned &rows, unsigned &cols)
{
	check_access(M_WRITE);

#ifdef WIN32
	// XXX Figure out how to do this
	rows = 24;
	cols = 80;
#else
	struct winsize wsz;
	ioctl(*handle, TIOCGWINSZ, &wsz);
	rows = wsz.ws_row;
	cols = wsz.ws_col;
#endif
}

unsigned Console::do_write(const char *buf, unsigned len)
{
	check_access(M_WRITE);

	return sys_write(handle, buf, len);
}

unsigned Console::do_read(char *buf, unsigned len)
{
	check_access(M_READ);

	unsigned ret = sys_read(handle, buf, len);
	if(ret==0)
		eof_flag = true;

	return ret;
}

Console &Console::instance(unsigned n)
{
	static Console in(0);
	static Console out(1);
	static Console err(2);

	switch(n)
	{
	case 0: return in;
	case 1: return out;
	case 2: return err;
	}

	throw invalid_argument("Console::instance");
}

Console &cin = Console::instance(0);
Console &cout = Console::instance(1);
Console &cerr = Console::instance(2);

} // namespace IO
} // namespace Msp
